#!/usr/bin/env python

"""
Parsing of vCalendar and iCalendar files.

Copyright (C) 2008 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free
Software Foundation; either version 3 of the License, or (at your option) any
later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
details.

You should have received a copy of the GNU Lesser General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.

--------

References:

RFC 2445: Internet Calendaring and Scheduling Core Object Specification
          (iCalendar)
          http://rfc.net/rfc2445.html
"""

import vContent

try:
    set
except NameError:
    from sets import Set as set

# Format details.

QUOTED_PARAMETERS = set([
    "ALTREP", "DELEGATED-FROM", "DELEGATED-TO", "DIR", "MEMBER", "SENT-BY"
    ])
MULTIVALUED_PARAMETERS = set([
    "DELEGATED-FROM", "DELEGATED-TO", "MEMBER"
    ])
QUOTED_TYPES = set(["URI"])

# Parser classes.

class vCalendarStreamParser(vContent.StreamParser):

    "A stream parser specifically for vCalendar/iCalendar."

    def next(self):

        """
        Return the next content item in the file as a tuple of the form
        (name, parameters, value).
        """

        name, parameters, value = vContent.StreamParser.next(self)
        return name, self.decode_parameters(parameters), self.decode_content(value)

    def decode_content(self, value):

        "Decode the given 'value', replacing quoted separator characters."

        # Replace quoted characters (see 4.3.11 in RFC 2445).

        value = vContent.StreamParser.decode_content(self, value)
        return value.replace("\\,", ",").replace("\\;", ";")

    # Internal methods.

    def decode_quoted_value(self, value):

        "Decode the given 'value', returning a list of decoded values."

        if value[0] == '"' and value[-1] == '"':
            return value[1:-1]
        else:
            return value

    def decode_parameters(self, parameters):

        """
        Decode the given 'parameters' according to the vCalendar specification.
        """

        decoded_parameters = {}

        for param_name, param_value in parameters.items():
            if param_name in QUOTED_PARAMETERS:
                param_value = self.decode_quoted_value(param_value)
                separator = '","'
            else:
                separator = ","
            if param_name in MULTIVALUED_PARAMETERS:
                param_value = param_value.split(separator)
            decoded_parameters[param_name] = param_value

        return decoded_parameters

class vCalendarParser(vContent.Parser):

    "A parser specifically for vCalendar/iCalendar."

    def parse(self, f, parser_cls=None):
        return vContent.Parser.parse(self, f, vCalendarStreamParser)

# Writer classes.

class vCalendarStreamWriter(vContent.StreamWriter):

    "A stream writer specifically for vCard."

    def write(self, name, parameters, value):

        """
        Write a content line for the given 'name', 'parameters' and 'value'
        information.
        """

        vContent.StreamWriter.write(self, name, self.encode_parameters(parameters), value)

    # Internal methods.

    def encode_quoted_value(self, value):

        "Encode the given 'value'."

        return '"%s"' % value

    def encode_parameters(self, parameters):

        """
        Encode the given 'parameters' according to the vCalendar specification.
        """

        encoded_parameters = {}

        for param_name, param_value in parameters.items():
            if param_name in QUOTED_PARAMETERS:
                param_value = self.encode_quoted_value(param_value)
                separator = '","'
            else:
                separator = ","
            if param_name in MULTIVALUED_PARAMETERS:
                param_value = separator.join(param_value)
            encoded_parameters[param_name] = param_value

        return encoded_parameters

    def encode_content(self, value):

        "Encode the given 'value', replacing separator characters."

        # Replace quoted characters (see 4.3.11 in RFC 2445).

        value = vContent.StreamWriter.encode_content(self, value)
        return value.replace(";", "\\;").replace(",", "\\,")

# Public functions.

def parse(f, non_standard_newline=0):

    """
    Parse the resource data found through the use of the file object 'f', which
    should provide Unicode data. (The codecs module can be used to open files or
    to wrap streams in order to provide Unicode data.)

    The optional 'non_standard_newline' can be set to a true value (unlike the
    default) in order to attempt to process files with CR as the end of line
    character.

    As a result of parsing the resource, the root node of the imported resource
    is returned.
    """

    return vContent.parse(f, non_standard_newline, vCalendarParser)

def iterparse(f, non_standard_newline=0):

    """
    Parse the resource data found through the use of the file object 'f', which
    should provide Unicode data. (The codecs module can be used to open files or
    to wrap streams in order to provide Unicode data.)

    The optional 'non_standard_newline' can be set to a true value (unlike the
    default) in order to attempt to process files with CR as the end of line
    character.

    An iterator is returned which provides event tuples describing parsing
    events of the form (name, parameters, value).
    """

    return vContent.iterparse(f, non_standard_newline, vCalendarStreamParser)

# vim: tabstop=4 expandtab shiftwidth=4
